#! /bin/bash

#  Utility functions used by the start_lepton.sh script

#---------------------------------------------------------------------
#  Compute start and end steps for the lepton daemon and the nodes
#
#  Input:
#
#  start_step     : the first step to be run by lepton (in s)
#  duration       : the duration of the simulation (in s), or -1
#  start_deferred : how long we must wait before starting the simulation (in s)
#  jitter         : a jitter value for starting nodes (in s)
#  in_hist        : the name of the file that contains the nodes history
#  nodes          : the number of nodes or a ',' separated list of 'profile:nb'
#
#  Output:
#
#  lepton_start_time : the time when lepton should run the 1st step (in ms)
#  start_step        : the first step to be run by lepton (in ms)
#  end_step          : the last step to be run by lepton (in ms), or -1
#  nodes_hist        : (node_id,node_start_step,node_end_step;)* with:
#      node_id         : simple id or id:profile
#      node_start_step : step when the node should be added (in ms)
#      node_end_step   : step when the node should be deleted (in ms), or -1
#---------------------------------------------------------------------
compute_simul_history() {

    if [ "${hist}" == "" ]; then
	
	#  Set the hist variable to ensure this is done only once
	hist="true"

	#  Compute the start_step in ms
	if [ "${start_step}" != "" ] && (( ${start_step} > 0 )); then
	    start_step=$((${start_step}*1000))
	else
	    start_step=0
	fi
	
	#  Compute the end_step in ms or -1 if no duration is defined
	end_step=-1
	if [ "${duration}" != "" ] && (( ${duration} > 0 )); then
	    end_step=$((${duration}*1000+${start_step}))
	fi

	#  Compute the history of the nodes
	if [ "${in_hist}" == "" ]; then     # no history
	    if [ "${in_dgs}" == "" ]; then  # no DGS input file
		compute_nodes_history
	    else                            # a DGS input file but no history
		hist="false"
	    fi
	fi

	#  Initialize the start_time value to the current time (in s)
	start_time=$(date +%s)

	#  Compute the absolute lepton start time in ms
	lepton_start_time=$((${start_time}*1000))
	    
	#  Add start_deferred if it is defined
	if [ "${start_deferred}" != "" ] && (( ${start_deferred} > 0 )); then
	    lepton_start_time=$((${start_deferred}*1000+${lepton_start_time}))
	fi
    fi
}

#---------------------------------------------------------------------
#  Compute start and end steps for the nodes from the 'nodes'
#  property
#
#  Input:
#
#  start_step        : The first step to be run by lepton (in ms)
#  end_step          : The last step to be run by lepton (in ms), or -1
#  jitter            : A jitter value for starting nodes (in s)
#  nodes             : The number of nodes or a ',' separated list of
#                      'profile:nb'
#
#  Output:
#
#  nodes_hist        : (node_start_step,node_end_step,node_id;)* with:
#      node_start_step : step when the node should be added (in ms)
#      node_end_step   : step when the node should be deleted (in ms), or -1
#      node_id         : simple id or id:profile
#---------------------------------------------------------------------
compute_nodes_history() {
		
    if [[ ${nodes} =~ ^[0-9]+$ ]]; then #  integer value (nb of nodes)
	
	nb_digits=${#nodes}
	if [ "${jitter}" != "" ] && (( ${jitter} > 0 )); then
	    randoms=$(get_randoms ${nodes} 1000 ${seed})
	fi
	for (( num_node = 0; num_node < ${nodes}; num_node++ )); do
	    node_id=$(create_node_ID ${num_node} ${nb_digits})
	    nodes_hist="${nodes_hist}$(make_node_history);"
	done
	
    else                                #  ',' separated list 'nb:profile'

	nb=$(get_nb_nodes ${nodes})
	nb_digits=${#nb}
	num_node=0
	if [ "${jitter}" != "" ] && (( ${jitter} > 0 )); then
	    randoms=$(get_randoms ${nb} 1000 ${seed})
	fi	
	for token in $(echo ${nodes} | sed 's/,/ /g') ; do
	    
	    nb_nodes=$(echo ${token} | cut -d: -f2)
	    node_profile=$(echo ${token} | cut -d: -f1)
	    
	    for (( i = 0; i < ${nb_nodes}; i++ )); do
		node_id=$(create_node_ID ${num_node} ${nb_digits})
		node_id="${node_id}:${node_profile}"
		nodes_hist="${nodes_hist}$(make_node_history);"
		num_node=$(($num_node+1))
	    done	    
	done
    fi
}

make_node_history() {
    node_start_step=$((${start_step}+$(get_random_delay)))
    node_duration=-1
    if (( ${end_step} > 0 )); then
	node_duration=$((${end_step}-${node_start_step}))
    fi
    echo "${node_start_step},${end_step},${node_duration},${node_id}"
}
	
#---------------------------------------------------------------------
#  Start the lepton daemon using the start_lepton function
#
#  Arguments: [info] [conf=confFile]* [key=value]*
#
#  Variables:
#
#  lepton_home  : the LEPTON install directory
#  log_dir      : the log directory where standard outputs are redirected
#---------------------------------------------------------------------
start_lepton_daemon() {

    #  Ensure that this host is supposed to run LEPTON
    if [ "${is_lepton_host}" == "true" ]; then

	if (( $# > 0 )) && [ "$1" == "info" ]; then
	    
	    #  List properties. No output redirection
	    show_config
	    start_lepton $*
	    
	else
	    
	    #  Create the log directory if it does not exist
	    if [ "${log_dir}" != "" ] && [ ! -d ${log_dir} ]; then
		mkdir -p ${log_dir}
	    fi

	    #  Information messages
	    echo "Redirecting outputs to ${log_dir}/lepton.out and ${log_dir}/lepton.err"
	    if [ "${out_dgs}" != "" ]; then
		echo "Writing the graph events to ${out_dgs}"
	    fi
	    
	    #  Start the lepton daemon
	    show_config > ${log_dir}/lepton.out
	    start_lepton $* >> ${log_dir}/lepton.out 2>  ${log_dir}/lepton.err &
	fi
    fi
}


#---------------------------------------------------------------------
#  Start the emulated nodes using the start_node function.
#
#  lepton_start_time : The time when lepton should run the 1st step (in ms)
#  start_step        : The first step to be run by lepton (in ms)
#  num_host          : index of the host where the script runs
#  nb_hosts          : the total number of hosts involved
#  time_margin       : how long we should start the node before it
#                      actually starts (in s)
#  nodes_hist        : (node_start_step,node_end_step,node_id;)* with:
#      node_start_step : step when the node should be added (in ms)
#      node_end_step   : step when the node should be deleted (in ms), or -1
#      node_id         : simple id or id:profile
#
#  $num_hosts and $nb_hosts allow to run concurrently on several hosts
#  (typically on a cluster). In that case each host will only create a
#  subset of the nodes (using a round-robin policy). If the simulation
#  is run on a single host, then $num_host=0 and $nb_hosts=1.
#---------------------------------------------------------------------
start_nodes() {

    start_step=${start_step:-0} 
    start=$((${lepton_start_time}-${start_step}))
    num_node=0

    if [ "${nodes_hist}" != "" ]; then
	
	for node_hist in $(echo ${nodes_hist} | sed 's/;/ /g') ; do
    
	    #  Ensure that this host is supposed to run this node
	    if (( ${num_host} == $((${num_node} % ${nb_hosts})) )); then
		
		start_a_node &
	    fi
	    
	    num_node=$((${num_node}+1))
	done

    elif [ "${in_hist}" != "" ]; then
	grep -v "^#" ${in_hist} | tr -s ' ' | \
            while read line
            do
    
		#  Ensure that this host is supposed to run this node
		if (( ${num_host} == $((${num_node} % ${nb_hosts})) )); then
		    
		    node_hist=$(echo ${line} | sed 's/ /,/g')		    
		    start_a_node &
		fi
	    
		num_node=$((${num_node}+1))
	    done
    fi
}

#---------------------------------------------------------------------
#  Start a single emulated node using the start_node function.
#
#  start           : the time when the step 0 occurs (in ms)
#  node_hist       : node_start_step,node_end_step,node_duration,node_id
#  num_node        : index of the node to be started
#  time_margin     : how long we should start the node before it
#                    actually starts (in s)
#
#  Each application node is created a short while before it is
#  actually started, according to $node_start_time and $start_margin
#  arguments. Thus, if node N is supposed to run from time
#  $node_start_time, then the corresponding process is created at
#  ($node_start_time - $start_margin), and this process is passed
#  argument $node_start_time and $node_end_time so it can start and
#  stop at the appropriate times.
#---------------------------------------------------------------------
start_a_node() {
    
    #  Get the node start/end steps
    node_id=$(echo ${node_hist} | cut -d, -f4)
    node_start_step=$(echo ${node_hist} | cut -d, -f1)
    node_end_step=$(echo ${node_hist} | cut -d, -f2)
		
    #  Compute the node start/end times
    node_start_time=$((${start}+${node_start_step}))
    node_end_time=-1
    if ((${node_end_step} > 0)); then
	node_end_time=$((${start}+${node_end_step}))
    fi

    #  Compute the node seed
    node_seed=$((${seed}+${num_node}*1000))
		
    #  Defer start until (node_start_time - time_margin)
    sleep_time=$((${node_start_time}/1000-${time_margin:-0}-$(date +%s)))
    if (( ${sleep_time} > 0 )); then
	sleep ${sleep_time}
    fi

    #  Call the start_node function
    start_node > ${log_dir}/${node_id%:*}.out 2>  ${log_dir}/${node_id%:*}.err 
}

#---------------------------------------------------------------------
#  Utility functions to generate informations about nodes
#---------------------------------------------------------------------
#  Create a node id in the form "Nxxxxx" where xxxxx are digits
#  Arguments: node number, number of digits
create_node_ID(){
    node_number=$1
    nb_digits=$2
    length=${#node_number}
    node_id="N";
    #  Add zeros
    for (( i = ${length}; i < $((${nb_digits})); i++ )); do
	node_id="${node_id}0"
    done
    #  Add the node number
    node_id="${node_id}${node_number}"
    echo ${node_id}
}

#  Give the total number of nodes
#  Argument: ',' separated list of 'profile:nb'
get_nb_nodes() {
    sum=0
    for token in $(echo $1 | sed 's/,/ /g') ; do
	nb=${token#*:}
	sum=$((${nb} + ${sum}))
    done
    echo ${sum}
}

#  Give a random delay in ms from a given jitter in s.
#  Variables: jitter, num_node, randoms
get_random_delay() {
    if [ "${randoms}" != "" ]; then
	random=$(echo ${randoms} | cut -d" " -f$(($num_node+1)))
	echo $((${random}*${jitter}))
    else
	echo 0
    fi	
}

#  Give random integers
#  Arguments: nb_values max_value base_seed
get_randoms() {
    seed=$3
    for (( i = 0; i < $1; i++ )); do
	seeds="${seeds} ${seed}"
	seed=$((${seed}+1000))
    done
    classpath=${lepton_home}/libs/casa-util-1.0.jar:${lepton_home}/libs/lepton-1.0.jar
    java -cp ${classpath} casa.util.randomizer $2 $seeds
}
